#
# This file is part of Config-Model-Tester
#
# This software is Copyright (c) 2013-2017 by Dominique Dumont.
#
# This is free software, licensed under:
#
#   The GNU Lesser General Public License, Version 2.1, February 1999
#

# This file was automatically generated by Dist::Zilla::Plugin::ModuleBuild v6.008.
use strict;
use warnings;

use Module::Build 0.34;


my %module_build_args = (
  "build_requires" => {
    "Module::Build" => "0.34"
  },
  "configure_requires" => {
    "Module::Build" => "0.34"
  },
  "dist_abstract" => "Test framework for Config::Model",
  "dist_author" => [
    "Dominique Dumont"
  ],
  "dist_name" => "Config-Model-Tester",
  "dist_version" => "2.061",
  "license" => "lgpl",
  "module_name" => "Config::Model::Tester",
  "recursive_test_files" => 1,
  "requires" => {
    "File::Copy::Recursive" => 0,
    "Log::Log4perl" => "1.11",
    "Path::Tiny" => 0,
    "Test::Differences" => 0,
    "Test::Exception" => 0,
    "Test::File::Contents" => 0,
    "Test::Memory::Cycle" => 0,
    "Test::More" => 0,
    "Test::Warn" => 0,
    "perl" => "5.010"
  }
);


my %fallback_build_requires = (
  "Module::Build" => "0.34"
);


unless ( eval { Module::Build->VERSION(0.4004) } ) {
  delete $module_build_args{test_requires};
  $module_build_args{build_requires} = \%fallback_build_requires;
}

my $build = Module::Build->new(%module_build_args);


$build->create_build_script;
